import gulp from 'gulp';
import babel from 'gulp-babel';
import babelCompiler from 'babel-register';
import gulpLoadPlugins from 'gulp-load-plugins';

import mocha from 'gulp-mocha';
import eslint from 'gulp-eslint';
import runSequence from 'run-sequence';
import gulpIf from 'gulp-if';
import * as isparta from 'isparta';


const plugins = gulpLoadPlugins();

const options = {
  codeCoverage: {
    reporters: ['lcov', 'text-summary'],
    thresholds: {
      global: { statements: 75, branches: 80, functions: 75, lines: 50 },
    },
  },
};

function isFixed(file) {
  // Has ESLint fixed the file contents?
  return file.eslint != null && file.eslint.fixed;
}

gulp.task('lint', () => {
  return gulp.src([
    'examples/**/*.js',
    'src/**/*.js',
    'test/**/*.js',
  ], { base: './' })
    .pipe(eslint())
    .pipe(eslint.format())
    .pipe(eslint.failAfterError());
});

gulp.task('lint-n-fix', () => {
  return gulp.src([
    'examples/**/*.js',
    'src/**/*.js',
    'test/**/*.js',
  ], { base: './' })
  .pipe(eslint({
    fix: true
  }))
  .pipe(eslint.format())
  // if fixed, write the file to dest
  .pipe(gulpIf(isFixed, gulp.dest('.')));
});

gulp.task('babel', () => {
  return gulp.src('src/**/*.js')
    .pipe(babel())
    .pipe(gulp.dest('dist'));
});

gulp.task('babel-examples', () => {
  return gulp.src('examples/**/*.js')
    .pipe(babel())
    .pipe(gulp.dest('distExamples'));
});

function runTests() {
  let reporters;

  if (plugins.util.env['code-coverage-reporter']) {
    reporters = [...options.codeCoverage.reporters, plugins.util.env['code-coverage-reporter']];
  } else {
    reporters = options.codeCoverage.reporters;
  }

  return gulp.src('test/**/*.js', { read: false })
    .pipe(mocha({
      reporter: process.env.MOCHA_REPORTER || 'spec',
      compilers: {
        js: babelCompiler
      }
    }))
    .on('error', err => {
      console.log(err);
      process.exit(1);
    })
    .pipe(plugins.istanbul.writeReports({
      dir: './coverage',
      reporters,
    }))
    // Enforce test coverage
    .pipe(plugins.istanbul.enforceThresholds({
      thresholds: options.codeCoverage.thresholds,
    }));
}
// covers files for code coverage
gulp.task('pre-test', () =>
  gulp.src(['src/**/*.js', '!gulpfile.babel.js'])
  // Covering files
  .pipe(plugins.istanbul({
    instrumenter: isparta.Instrumenter, // eslint-disable-line no-undef
    includeUntested: true,
  }))
  // Force `require` to return covered files
  .pipe(plugins.istanbul.hookRequire())
);

gulp.task('test', ['pre-test'], () => runTests());

gulp.task('build', done => runSequence('babel', 'babel-examples', 'lint', 'test', done));
