/**
 * Created by asafdavid on 07/12/2015.
 */
// External packages
import amqp from 'amqplib';
import Chance from 'chance';
import _ from 'lodash';
import { EventEmitter } from 'events';

import logger from './logger';

const RECONNECTION_TIMEOUT = 1000;
const RETRY_CONNECT_TIMEOUT = 5000;
const chance = new Chance();

/**
 * Represent a connection to the AMQP server
 */
export default class Connection extends EventEmitter {
  /**
   * Creates a new connection object
   * @param connectionString
   */
  constructor(connectionString) {
    super();
    this.connectionString = _.castArray(connectionString || 'amqp://localhost');
  }

  /**
   * Returns a random connection string from the pool
   * @returns {*}
   */
  getConnectionString() {
    return chance.pickone(this.connectionString);
  }

  /**
   * Connects to the wanted AMQP server and tries to maintain the connection.
   */
  connect() {
    const connectionString = this.getConnectionString();
    logger.instance.info(`[AMQP] picked a connection string ${connectionString}`);
    return amqp.connect(connectionString).then((conn) => {
      this.conn = conn;

      logger.instance.info(`[AMQP] connected to ${connectionString}`);

      // Opens a channel
      conn.createConfirmChannel().then((ch) => {
        this.channel = ch;
        this.emit('open', conn);

        this.channel.on('error', (err) => {
          logger.instance.error('[AMQP] error occurred in channel', { error: err });
        });
      });

      // Handle connection error
      conn.on('error', (err) => {
        if (err.message === 'read ECONNRESET') {
          logger.instance.error('[AMQP] reconnecting');
          setTimeout(() => this.connect(), RECONNECTION_TIMEOUT);
        } else {
          logger.instance.error('[AMQP] conn error', { error: err });
          this.emit('error', err);
        }
      });

      conn.on('close', (msg) => {
        if (!!msg && msg !== 'Closed by client' && !msg.match(/Error/)) {
          logger.instance.error('[AMQP] reconnecting');
          setTimeout(() => this.connect(), RECONNECTION_TIMEOUT);
        } else {
          this.emit('close');
        }
      });
    }, (err) => {
      logger.instance.error('[AMQP] conn error', { error: err });
      if (err.message.match(/ECONNREFUSED|ENOTFOUND/)) {
        logger.instance.error('[AMQP] reconnecting');
        setTimeout(() => this.connect(), RETRY_CONNECT_TIMEOUT);
      } else {
        this.emit('error', err);
      }
    });
  }

  close() {
    logger.instance.info('[AMQP] Closing channel and connection');
    return this.channel.waitForConfirms()
      .then(() => this.channel.close())
      .then(() => this.conn.close());
  }
}
