/**
 * Created by asafdavid on 23/02/2017.
 */
import _ from 'lodash';
import Promise from 'bluebird';
import * as Errio from 'errio';

import Logger from '../lib/logger';
import Router from '../lib/handlers/router';
import BaseService from './baseService';
import Client from '../lib/client';

const Log = Logger.instance;

export default class RabbitService {
  /**
   * Registers a service to a router
   * @param router
   * @param service
   */
  static register(router, service, tasks = []) {
    // Verify parameters
    if (!(router && router instanceof Router)) {
      throw new TypeError('router has to be an instance of Router');
    }
    if (!(service &&
      (service instanceof BaseService || service.prototype instanceof BaseService))) {
      throw new TypeError('service has to be an instance of BaseService');
    }

    // Extract handlers
    const handlers = RabbitService._getTasks(service, tasks);
    router.addRoutes(RabbitService.toRabbitHandlers(handlers));
  }

  /**
   * Placeholder to convert handlers to rabbit syntax
   * @param handlers {Object}
   * @returns {*}
   */
  static toRabbitHandlers(tasks) {
    const handlers = [];
    _.each(tasks, (task) => {
      const { name, topic, handler, middleware, options } = task;

      handlers.push({
        name,
        pattern: topic,
        middleware,
        handle: ({ content, meta }) => handler(content, meta),
        options,
      });
    });

    return handlers;
  }

  /**
   * Returns a send interface for a specific task
   * @param serviceName
   * @param taskName
   * @param client
   * @param topic
   * @param baseError
   * @param serverMeta
   * @param sync
   * @returns {*}
   */
  static getTaskClientInterface(
    serviceName,
    taskName,
    client,
    topic,
    baseError,
    serverMeta,
    sync,
    timeout) {
    return sync ? function interfaceSendSync(msg, meta) {
      return RabbitService.sendSync(
        serviceName,
        taskName,
        client,
        serverMeta,
        topic,
        baseError,
        msg,
        meta,
        timeout
      );
    } : function interfaceSendAsync(msg, meta, context) {
      return RabbitService.sendAsync(
        serviceName,
        taskName,
        client,
        serverMeta,
        topic,
        msg,
        meta,
        context,
      );
    };
  }

  /**
   * @param serviceName
   * @param client
   * @param sync
   * @param serverMeta
   * @param [mapping] A mapping object that can be used in order to create methods
   * that wrap the dispatch function, with a specific topic.
   * should be in the form: { taskName: topic }
   * @returns {{}}
   */
  static getGenericClient(serviceName, client, sync, serverMeta, mapping) {
    const genericClient = {};

    genericClient.dispatch = sync ? function genericClientSync(topic, msg, meta) {
      return RabbitService.sendSync(
        serviceName,
        topic,
        client,
        serverMeta,
        topic,
        null,
        msg,
        meta,
      );
    } : function genericClientAsync(topic, msg, meta, context) {
      return RabbitService.sendAsync(
        serviceName,
        topic,
        client,
        serverMeta,
        topic,
        msg,
        meta,
        context,
      );
    };

    // if a mapping object was supplied, create the wrapper methods
    if (mapping) {
      _.mapKeys(mapping, (topic, taskName) => {
        genericClient[taskName] = (msg, meta, context) =>
          genericClient.dispatch(topic, msg, meta, context);
      });
    }

    return genericClient;
  }

  /**
   * Returns a client to execute the the service's tasks.
   * @param service
   * @param client
   * @param serverMeta
   * @param tasks
   */
  static getClient(service, client, serverMeta, tasks = []) {
    // Validation
    if (!(service &&
      (service instanceof BaseService || service.prototype instanceof BaseService))) {
      throw new TypeError('service has to be an instance of BaseService');
    }
    if (!(client instanceof Client)) throw new TypeError('client must be an instance of Client');

    if (!(serverMeta && (serverMeta.serverExchangeName || serverMeta.serverQueueName))) {
      throw new TypeError('wrong serverMeta was provided');
    }

    // Build the client object
    const serviceName = service.prototype.constructor.name;
    const wantedTasks = RabbitService._getTasks(service, tasks);
    const clientInterface = {};
    // register class level errors
    const { _errors, _baseError } = service.prototype.constructor;
    const classBaseError = _baseError;
    const classErrors = _errors;
    RabbitService._registerErrors(classErrors, classBaseError);

    _.each(wantedTasks, (task) => {
      const { name, topic, sync, errors, baseError, timeout } = task;
      // register task lever errors
      RabbitService._registerErrors(errors, baseError);

      clientInterface[name] = RabbitService.getTaskClientInterface(
        serviceName,
        name,
        client,
        topic,
        baseError || classBaseError,
        serverMeta,
        sync,
        timeout,
      );
    });

    return clientInterface;
  }

  /**
   * Returns all of the tasks for a specific service
   * @param service
   * @param tasks
   * @returns {*}
   * @private
   */
  static _getTasks(service, tasks) {
    const allTasks = service._tasks || service.prototype._tasks;
    return tasks && !_.isEmpty(tasks) ? _.pick(allTasks, tasks) : allTasks;
  }

  static _registerErrors(errors, baseError) {
    _.each(errors, (error) => {
      Errio.register(error);
    });
    if (baseError) Errio.register(baseError);
  }

  /**
   * Executes an async task
   * @param serviceName
   * @param taskName
   * @param client
   * @param serverMeta
   * @param topicName
   * @param data
   * @param context
   * @returns {Promise.<T>}
   */
  static sendAsync(serviceName, taskName, client, serverMeta, topicName, data, meta, context) {
    return client.sendAsync(serverMeta, topicName, data, meta, context)
      .catch(() => {
        const error = `could not send message to ${topicName}`;
        Log.error(`[${serviceName}/${taskName}] - ${error}`);
        return Promise.reject(new Error(error));
      });
  }

  /**
   * Executes a sync task
   * @param serviceName
   * @param taskName
   * @param client
   * @param serverMeta
   * @param topicName
   * @param baseError
   * @param data
   * @param meta
   * @param timeout
   * @returns {Promise.<T>}
   */
  static sendSync(
    serviceName,
    taskName,
    client,
    serverMeta,
    topicName,
    baseError,
    data,
    meta,
    timeout
  ) {
    return client.sendSync(serverMeta, topicName, data, meta, timeout)
      .then((result) => {
        const content = result && result.content.toString();
        return content && JSON.parse(content);
      })
      .catch((error) => {
        if (error instanceof Promise.TimeoutError) {
          return Promise.reject(new Promise.TimeoutError());
        }

        const errorString = error.content ? error.content.toString() : error.toString();
        error = Errio.parse(errorString);

        // if base error is set and the error isn't an instance of base error
        if (baseError && !(error instanceof baseError)) {
          error = new baseError(error.toString()); // eslint-disable-line new-cap
        }

        // Store metadata on the error to identify the service/task
        error.meta = {
          service: serviceName,
          task: taskName,
        };

        Log.error(`[${serviceName}/${taskName}] - error for topic ${topicName}: ${error}`);
        return Promise.reject(error);
      });
  }
}
